import copy
import yaml
import subprocess
import os.path

import six

# largely copied from my earlier work in fakemtpd


if hasattr(subprocess, 'check_output'):
    check_output = subprocess.check_output
else:
    def check_output(*args, **kwargs):
        kwargs['stdout'] = subprocess.PIPE
        p = subprocess.Popen(*args, **kwargs)
        stdout, _ = p.communicate()
        assert p.returncode == 0
        return stdout


def _param_getter_factory(parameter):
    def f(self):
        return self._config[parameter]
    f.__name__ = parameter
    return f


class _ParamsAsProps(type):
    """Create properties on the classes that apply this for everything
    in cls._parameters which read out of self._config.

    Cool fact: you can override any of these properties by just defining
    your own with the same name. Just like if they were statically defined!"""
    def __new__(clsarg, name, bases, d):
        cls = super(_ParamsAsProps, clsarg).__new__(clsarg, name, bases, d)
        for parameter in cls._parameters.keys():
            if parameter not in d:
                f = _param_getter_factory(parameter)
                setattr(cls, parameter, property(f))
        return cls


class ConfigError(Exception):
    def __init__(self, message):
        self.message = message

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.message)

    def __str__(self):
        return '%s(%r)' % (self.__class__.__name__, self.message)


@six.add_metaclass(_ParamsAsProps)
class Config(object):
    _parameters = {
        'smtp_host': '127.0.0.1',
        'smtp_port': 25,
        'smtp_ssl': True,  # if false, do STARTTLS
        'smtp_username': '',
        'smtp_password': None,
        'smtp_password_command': None,
        'smtp_timeout': 10,
        'css_file': None,
        'sendmail': '/usr/sbin/sendmail',
    }

    def __init__(self):
        self._config = copy.copy(self._parameters)
        self._css = None

    def merge_config(self, d):
        invalid_keys = set(d.keys()) - set(self._config.keys())
        if invalid_keys:
            raise ConfigError('Unexpected config keys: %s' % ', '.join(sorted(invalid_keys)))
        for key in self._config:
            if key in d:
                self._config[key] = d[key]
        if self._config['smtp_password'] and self._config['smtp_password_command']:
            raise ConfigError('Cannot set smtp_password *and* smtp_password_command')
        if self._config['css_file']:
            self._css = None
            self._config['css_file'] = os.path.expanduser(self._config['css_file'])
            if not os.path.exists(self._config['css_file']):
                raise ConfigError('CSS file %s does not exist' % self._config['css_file'])

    def load(self, fobj):
        d = yaml.safe_load(fobj)
        self.merge_config(d)

    @property
    def css(self):
        if self._css is None:
            if self.css_file is not None:
                with open(os.path.expanduser(self.css_file), 'r') as f:
                    self._css = f.read()
            else:
                self._css = ''
        return self._css

    @property
    def smtp_password(self):
        if self._config['smtp_password_command']:
            return check_output(self._config['smtp_password_command'], shell=True, universal_newlines=True).rstrip('\n')
        else:
            return self._config['smtp_password']



import tempfile

from muttdown.config import Config


def test_smtp_password_literal():
    c = Config()
    c.merge_config({'smtp_password': 'foo'})
    assert c.smtp_password == 'foo'


def test_smtp_password_command():
    c = Config()
    c.merge_config({'smtp_password_command': 'sh -c "echo foo"'})
    assert c.smtp_password == 'foo'


def test_css():
    c = Config()
    c.merge_config({'css_file': None})
    assert c.css == ''

    with tempfile.NamedTemporaryFile(delete=True) as css_file:
        css_file.write(b'html { background-color: black; }\n')
        css_file.flush()
        c.merge_config({'css_file': css_file.name})
        assert c.css == 'html { background-color: black; }\n'



from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication
from email.message import Message
import tempfile
import shutil
import socket
import threading
import ssl
import select
import time
import os
import sys

import pytest
import yaml

from muttdown import main
from muttdown.main import convert_tree
from muttdown.main import process_message
from muttdown.config import Config


@pytest.fixture
def basic_config():
    return Config()


@pytest.fixture
def tempdir():
    # workaround because pytest's bultin tmpdir fixture is broken on python 3.3
    dirname = tempfile.mkdtemp()
    try:
        yield dirname
    finally:
        shutil.rmtree(dirname)


@pytest.fixture
def config_with_css(tempdir):
    with open('%s/test.css' % tempdir, 'w') as f:
        f.write('html, body, p { font-family: serif; }\n')
    c = Config()
    c.merge_config({'css_file': '%s/test.css' % tempdir})
    return c


def test_unmodified_no_match(basic_config):
    msg = Message()
    msg['Subject'] = 'Test Message'
    msg['From'] = 'from@example.com'
    msg['To'] = 'to@example.com'
    msg['Bcc'] = 'bananas'
    msg.set_payload('This message has no sigil')

    converted = process_message(msg, basic_config)
    assert converted == msg


def test_simple_message(basic_config):
    msg = MIMEMultipart()
    msg['Subject'] = 'Test Message'
    msg['From'] = 'from@example.com'
    msg['To'] = 'to@example.com'
    msg['Bcc'] = 'bananas'
    msg.preamble = 'Outer preamble'

    msg.attach(MIMEText("!m This is the main message body"))

    attachment = MIMEText('this is an attachment', 'x-misc')
    attachment.add_header('Content-Disposition', 'attachment')
    msg.attach(attachment)

    converted, _ = convert_tree(msg, basic_config)
    assert converted['Subject'] == 'Test Message'
    assert converted['From'] == 'from@example.com'
    assert converted['To'] == 'to@example.com'
    assert converted.get('Bcc', None) is None
    assert isinstance(converted, MIMEMultipart)
    assert converted.preamble == 'Outer preamble'
    assert len(converted.get_payload()) == 2
    alternatives_part = converted.get_payload()[0]
    assert isinstance(alternatives_part, MIMEMultipart)
    assert alternatives_part.get_content_type() == 'multipart/alternative'
    assert len(alternatives_part.get_payload()) == 2
    text_part = alternatives_part.get_payload()[0]
    html_part = alternatives_part.get_payload()[1]
    assert isinstance(text_part, MIMEText)
    assert text_part.get_content_type() == 'text/plain'
    assert isinstance(html_part, MIMEText)
    assert html_part.get_content_type() == 'text/html'
    attachment_part = converted.get_payload()[1]
    assert isinstance(attachment_part, MIMEText)
    assert attachment_part['Content-Disposition'] == 'attachment'
    assert attachment_part.get_content_type() == 'text/x-misc'


def test_with_css(config_with_css):
    msg = Message()
    msg['Subject'] = 'Test Message'
    msg['From'] = 'from@example.com'
    msg['To'] = 'to@example.com'
    msg['Bcc'] = 'bananas'
    msg.set_payload('!m\n\nThis is a message')

    converted, _ = convert_tree(msg, config_with_css)
    assert isinstance(converted, MIMEMultipart)
    assert len(converted.get_payload()) == 2
    text_part = converted.get_payload()[0]
    assert text_part.get_payload(decode=True) == b'!m\n\nThis is a message'
    html_part = converted.get_payload()[1]
    assert html_part.get_payload(decode=True) == b'<p style="font-family: serif">This is a message</p>'


def test_headers_when_multipart_signed(basic_config):
    msg = MIMEMultipart('signed')
    msg['Subject'] = 'Test Message'
    msg['From'] = 'from@example.com'
    msg['To'] = 'to@example.com'
    msg['Bcc'] = 'bananas'
    msg.preamble = 'Outer preamble'

    msg.attach(MIMEText("!m This is the main message body"))
    msg.attach(MIMEApplication('signature here', 'pgp-signature', name='signature.asc'))

    converted, _ = convert_tree(msg, basic_config)

    assert converted['Subject'] == 'Test Message'
    assert converted['From'] == 'from@example.com'
    assert converted['To'] == 'to@example.com'

    assert isinstance(converted, MIMEMultipart)
    assert converted.preamble == 'Outer preamble'
    assert len(converted.get_payload()) == 2
    assert converted.get_content_type() == 'multipart/alternative'
    html_part = converted.get_payload()[0]
    original_signed_part = converted.get_payload()[1]
    assert isinstance(html_part, MIMEText)
    assert html_part.get_content_type() == 'text/html'
    assert isinstance(original_signed_part, MIMEMultipart)
    assert original_signed_part.get_content_type() == 'multipart/signed'
    assert original_signed_part['Subject'] is None
    text_part = original_signed_part.get_payload()[0]
    signature_part = original_signed_part.get_payload()[1]
    assert text_part.get_content_type() == 'text/plain'
    assert signature_part.get_content_type() == 'application/pgp-signature'


class MockSmtpServer(object):
    def __init__(self):
        self._s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._s.bind(('127.0.0.1', 0))
        self.address = self._s.getsockname()[0:2]
        self._t = None
        self._started = threading.Event()
        self.messages = []
        self.running = False

    def start(self):
        self._t = threading.Thread(target=self.run)
        self._t.start()
        if self._started.wait(5) is not True:
            raise ValueError('SMTP Server Thread failed to start!')

    def run(self):
        if hasattr(ssl, 'create_default_context'):
            context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        else:
            context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
        context.load_cert_chain(certfile='tests/data/cert.pem', keyfile='tests/data/key.pem')
        self._s.listen(128)
        self._started.set()
        self.running = True
        while self.running:
            r, _, x = select.select([self._s], [self._s], [self._s], 0.5)
            if r:
                start = time.time()
                conn, addr = self._s.accept()
                conn = context.wrap_socket(conn, server_side=True)
                message = b''
                conn.sendall(b'220 localhost SMTP Fake\r\n')
                message += conn.recv(1024)
                conn.sendall(b'250-localhost\r\n250 DSN\r\n')
                # MAIL FROM
                message += conn.recv(1024)
                conn.sendall(b'250 2.1.0 Ok\r\n')
                # RCPT TO
                message += conn.recv(1024)
                conn.sendall(b'250 2.1.0 Ok\r\n')
                # DATA
                message += conn.recv(6)
                conn.sendall(b'354 End data with <CR><LF>.<CR><LF>\r\n')
                while time.time() < start + 5:
                    chunk = conn.recv(4096)
                    if not chunk:
                        break
                    message += chunk
                    if b'\r\n.\r\n' in message:
                        break
                conn.sendall(b'250 2.1.0 Ok\r\n')
                message += conn.recv(1024)
                conn.sendall(b'221 Bye\r\n')
                conn.close()
                self.messages.append((addr, message))

    def stop(self):
        if self._t is not None:
            self.running = False
            self._t.join()


@pytest.fixture
def smtp_server():
    s = MockSmtpServer()
    s.start()
    try:
        yield s
    finally:
        s.stop()


def test_main_smtplib(tempdir, smtp_server, mocker):
    config_path = os.path.join(tempdir, 'config.yaml')
    with open(config_path, 'w') as f:
        yaml.dump({
            'smtp_host': smtp_server.address[0],
            'smtp_port': smtp_server.address[1],
            'smtp_ssl': True
        }, f)
    msg = Message()
    msg['Subject'] = 'Test Message'
    msg['From'] = 'from@example.com'
    msg['To'] = 'to@example.com'
    msg['Bcc'] = 'bananas'
    msg.set_payload('This message has no sigil')
    mocker.patch.object(main, 'read_message', return_value=msg.as_string())
    main.main(['-c', config_path, '-f', 'from@example.com', 'to@example.com'])

    assert len(smtp_server.messages) == 1
    attr, transcript = smtp_server.messages[0]
    assert b'Subject: Test Message' in transcript
    assert b'no sigil' in transcript


def test_main_passthru(tempdir, mocker):
    output_path = os.path.join(tempdir, 'output')
    sendmail_path = os.path.join(tempdir, 'sendmail')
    with open(sendmail_path, 'w') as f:
        f.write('#!{0}\n'.format(sys.executable))
        f.write('import sys\n')
        f.write('output_path = "{0}"\n'.format(output_path))
        f.write('open(output_path, "w").write(sys.stdin.read())\n')
        f.write('sys.exit(0)')
    os.chmod(sendmail_path, 0o750)
    config_path = os.path.join(tempdir, 'config.yaml')
    with open(config_path, 'w') as f:
        yaml.dump({
            'sendmail': sendmail_path
        }, f)

    msg = Message()
    msg['Subject'] = 'Test Message'
    msg['From'] = 'from@example.com'
    msg['To'] = 'to@example.com'
    msg['Bcc'] = 'bananas'
    msg.set_payload('This message has no sigil')
    mocker.patch.object(main, 'read_message', return_value=msg.as_string())
    main.main(['-c', config_path, '-f', 'from@example.com', '-s', 'to@example.com'])

    with open(output_path, 'rb') as f:
        transcript = f.read()
    assert b'Subject: Test Message' in transcript
    assert b'no sigil' in transcript



#!/usr/bin/env python
from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    long_description = f.read()


setup(
    name="muttdown",
    version="0.3.4",
    author="James Brown",
    author_email="Roguelazer@gmail.com",
    url="https://github.com/Roguelazer/muttdown",
    license="ISC",
    packages=find_packages(exclude=['tests']),
    keywords=["email"],
    description="Sendmail replacement that compiles markdown into HTML",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        'Markdown>=2.0',
        'PyYAML>=3.0',
        'pynliner==0.8.0',
        'six',
    ],
    entry_points={
        'console_scripts': [
            'muttdown = muttdown.main:main',
        ]
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Intended Audience :: End Users/Desktop",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: ISC License (ISCL)",
        "Topic :: Communications :: Email",
    ]
)



from __future__ import print_function

import argparse
import sys
import smtplib
import re
import os.path
import email
import email.iterators
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

import subprocess
import six

import markdown
import pynliner

from . import config
from . import __version__

__name__ = 'muttdown'


def convert_one(part, config):
    text = part.get_payload(decode=True)
    if not isinstance(text, six.text_type):
        # no, I don't know why decode=True sometimes fails to decode.
        text = text.decode('utf-8')
    if not text.startswith('!m'):
        return None
    text = re.sub('\s*!m\s*', '', text, re.M)
    if '\n-- \n' in text:
        pre_signature, signature = text.split('\n-- \n')
        md = markdown.markdown(pre_signature, output_format="html5")
        md += '\n<div class="signature" style="font-size: small"><p>-- <br />'
        md += '<br />'.join(signature.split('\n'))
        md += '</p></div>'
    else:
        md = markdown.markdown(text)
    if config.css:
        md = '<style>' + config.css + '</style>' + md
        md = pynliner.fromString(md)
    message = MIMEText(md, 'html', _charset="UTF-8")
    return message


def _move_headers(source, dest):
    for k, v in source.items():
        # mutt sometimes sticks in a fake bcc header
        if k.lower() == 'bcc':
            del source[k]
        elif not (k.startswith('Content-') or k.startswith('MIME')):
            dest.add_header(k, v)
            del source[k]


def convert_tree(message, config, indent=0, wrap_alternative=True):
    """Recursively convert a potentially-multipart tree.

    Returns a tuple of (the converted tree, whether any markdown was found)
    """
    ct = message.get_content_type()
    cs = message.get_content_subtype()
    if not message.is_multipart():
        # we're on a leaf
        converted = None
        disposition = message.get('Content-Disposition', 'inline')
        if disposition == 'inline' and ct in ('text/plain', 'text/markdown'):
            converted = convert_one(message, config)
        if converted is not None:
            if wrap_alternative:
                new_tree = MIMEMultipart('alternative')
                _move_headers(message, new_tree)
                new_tree.attach(message)
                new_tree.attach(converted)
                return new_tree, True
            else:
                return converted, True
        return message, False
    else:
        if ct == 'multipart/signed':
            # if this is a multipart/signed message, then let's just
            # recurse into the non-signature part
            new_root = MIMEMultipart('alternative')
            if message.preamble:
                new_root.preamble = message.preamble
            _move_headers(message, new_root)
            converted = None
            for part in message.get_payload():
                if part.get_content_type() != 'application/pgp-signature':
                    converted, did_conversion = convert_tree(part, config, indent=indent + 1,
                                                             wrap_alternative=False)
                    if did_conversion:
                        new_root.attach(converted)
            new_root.attach(message)
            return new_root, did_conversion
        else:
            did_conversion = False
            new_root = MIMEMultipart(cs, message.get_charset())
            if message.preamble:
                new_root.preamble = message.preamble
            _move_headers(message, new_root)
            for part in message.get_payload():
                part, did_this_conversion = convert_tree(part, config, indent=indent + 1)
                did_conversion |= did_this_conversion
                new_root.attach(part)
            return new_root, did_conversion


def process_message(mail, config):
    converted, did_any_markdown = convert_tree(mail, config)
    if 'Bcc' in converted:
        del converted['Bcc']
    return converted


def smtp_connection(c):
    """Create an SMTP connection from a Config object"""
    if c.smtp_ssl:
        klass = smtplib.SMTP_SSL
    else:
        klass = smtplib.SMTP
    conn = klass(c.smtp_host, c.smtp_port, timeout=c.smtp_timeout)
    if not c.smtp_ssl:
        conn.ehlo()
        conn.starttls()
        conn.ehlo()
    if c.smtp_username:
        conn.login(c.smtp_username, c.smtp_password)
    return conn


def read_message():
    return sys.stdin.read()


def main(argv=None):
    parser = argparse.ArgumentParser(prog='muttdown')
    parser.add_argument('-v', '--version', action='version', version='%s %s' % (__name__, __version__))
    parser.add_argument(
        '-c', '--config_file', default=os.path.expanduser('~/.muttdown.yaml'),
        type=argparse.FileType('r'), required=False,
        help='Path to YAML config file (default %(default)s)'
    )
    parser.add_argument(
        '-p', '--print-message', action='store_true',
        help='Print the translated message to stdout instead of sending it'
    )
    parser.add_argument('-f', '--envelope-from', required=True)
    parser.add_argument(
        '-s', '--sendmail-passthru', action='store_true',
        help='Pass mail through to sendmail for delivery'
    )
    parser.add_argument('addresses', nargs='+')
    args = parser.parse_args(argv)

    c = config.Config()
    try:
        c.load(args.config_file)
    except config.ConfigError as e:
        sys.stderr.write('Error(s) in configuration %s:\n' % args.config_file.name)
        sys.stderr.write(' - %s\n' % e.message)
        sys.stderr.flush()
        return 1

    message = read_message()

    mail = email.message_from_string(message)

    rebuilt = process_message(mail, c)
    rebuilt.set_unixfrom(args.envelope_from)

    if args.print_message:
        print(rebuilt.as_string())
    elif args.sendmail_passthru:
        cmd = c.sendmail.split() + ['-f', args.envelope_from] + args.addresses

        proc = subprocess.Popen(cmd, stdin=subprocess.PIPE, shell=False)
        msg = rebuilt.as_string()
        if sys.version_info > (3, 0):
            msg = msg.encode('utf-8')
        proc.stdin.write(msg)
        proc.stdin.close()
        proc.wait()
        return proc.returncode
    else:
        conn = smtp_connection(c)
        conn.sendmail(args.envelope_from, args.addresses, rebuilt.as_string())
        conn.quit()
    return 0


if __name__ == '__main__':
    sys.exit(main())
